# ./orbitals/mo.py

import os, sys, copy, glob, ntpath
import shutil, json
import numpy as np
from chemscripts.calc import queue
from chemscripts.mylogging import createLogger
from chemscripts.nbo import NBO3LogParser, highest_p_character, generate_nbo_cube, generate_isosurface
from chemscripts import utils, geom
from chemscripts import nbo

STARTFILE = "./meoome.log" # The name of optimization log-file
CALCDIR = "./calc_files" # Dir for all new files
MO_IDXS =  ["LUMO+1", "LUMO", "HOMO", "HOMO-1", "HOMO-2", "HOMO-3"] # Desired MOs
ISOVALUE = 0.08 # Isovalue for all MOs
ASSIGNMENT_FNAME = "./mo_assignment.json"

def get_n_homo(mo_idxs):
    res = 0
    for item in mo_idxs: # item is either "LUMO*" (skip), "HOMO" (skip), or "HOMO-N" (get N)
        if "HOMO" not in item or "-" not in item:
            continue
        cur_n = int(item.split("-")[1])
        if cur_n > res:
            res = cur_n
    return res

def get_n_lumo(mo_idxs):
    res = 0
    for item in mo_idxs: # item is either "HOMO*" (skip), "LUMO" (skip), or "LUMO+N" (get N)
        if "LUMO" not in item or "+" not in item:
            continue
        cur_n = int(item.split("+")[1])
        if cur_n > res:
            res = cur_n
    return res

if __name__ == "__main__":
    for exe in ("rung", "formchk", "cubegen", "NBO6"):
        assert utils.check_availability(exe), "%s is not included in PATH" % exe

    if len(sys.argv) > 1:
        maxproc = int(sys.argv[1])
    else:
        maxproc = None
    gdriver = queue.init_thread(maxproc)
    logger = createLogger("Main")
    utils.checkout_directory(CALCDIR)
    start_fname = STARTFILE

    logger.info("Step 1. Run SCF calculation for initial geometry")
    logger.info("Processing " + start_fname)
    scfgjfs = []
    xyzs, syms = utils.parse_gaussian_log(start_fname)
    scf_gjfname = os.path.join(CALCDIR, ntpath.basename(start_fname).replace('.log', '_scf.gjf'))
    scf_checkpoint = ntpath.basename(scf_gjfname).replace('.gjf', '.chk')
    utils.write_gjf(xyzs, syms, 'singlepoint', scf_gjfname,
                    subs={
                        'nproc': 1,
                        'chkname': scf_checkpoint,
                    })
    scfgjfs.append(scf_gjfname)
    utils.start_calcs(scfgjfs, gdriver)
    logger.info("Step 1 is finished")

    logger.info("Step 2. Run NBO3 calculation")
    logger.info("Processing " + start_fname)
    nbo3gjfs = []
    nbo3_gjfname = os.path.join(CALCDIR, ntpath.basename(start_fname).replace('.log', '_nbo3.gjf'))
    utils.write_gjf(xyzs, syms, 'nborestart', nbo3_gjfname,
                    subs={
                        'nproc': 1,
                        'chkname': ntpath.basename(nbo3_gjfname).replace('.gjf', '.chk'),
                        'oldchkname': scf_checkpoint,
                        'file': ntpath.basename(start_fname).split('.')[0],
                    })
    nbo3gjfs.append(nbo3_gjfname)
    utils.start_calcs(nbo3gjfs, gdriver)
    logger.info("Step 2 is finished")

    logger.info("Step 3. Preparing for NBO6 analysis")
    nbo6_inpfiles = []
    for fsfile in glob.glob(os.path.join(CALCDIR, "*.47")):
        newname = fsfile.lower()
        shutil.move(fsfile, newname)
        lines = open(newname, 'r').readlines()
        lines[1] = ' $NBO DMNBO=W420 FNBO=W69 DMMO=W419 FMO=W699 file=%s $END\n' % ntpath.basename(newname).split('.')[0]
        with open(newname, 'w') as f:
            f.write(''.join(lines))
        nbo6_inpfiles.append(newname)
    logger.info("Step 3a. Running NBO6 analysis")
    utils.start_calcs(nbo6_inpfiles, gdriver)
    for file in glob.glob(os.path.join(CALCDIR, "*_NBO.out")):
        shutil.move(file, file.replace('_NBO.out', '_nbo6.out'))
    logger.info("Step 3 is finished")

    logger.info("Step 4. Organizing the obtained files")
    logger.info("Step 4a. Converting chk to fchk")
    fchkcommands = []
    for chkfile in glob.glob(os.path.join(CALCDIR, "*.chk")):
        fchkcommands.append(utils.get_formchk_call(chkfile))
    utils.start_calcs(fchkcommands, gdriver)
    logger.info("Step 4b. Renaming matrix files")
    for file in glob.glob(os.path.join(CALCDIR, "*.69")):
        os.remove(file)
        #shutil.move(file, file.replace(".69", "_nbo6.fnbo"))
    for file in glob.glob(os.path.join(CALCDIR, "*.699")):
        shutil.move(file, file.replace(".699", "_nbo6.fmo"))
    for file in glob.glob(os.path.join(CALCDIR, "*.420")):
        os.remove(file)
        #shutil.move(file, file.replace(".420", "_nbo6.dmnbo"))
    for file in glob.glob(os.path.join(CALCDIR, "*.419")):
        shutil.move(file, file.replace(".419", "_nbo6.dmmo"))
    logger.info("Step 4 is finished")

    logger.info("Step 5. Getting boundary MO data")
    orb_data = []
    N_HOMO = get_n_homo(MO_IDXS) # Number of orbitals below HOMO that we need to process
    N_LUMO = get_n_lumo(MO_IDXS) # Number of orbitals above LUMO that we need to process
    for nbo6out in glob.glob(os.path.join(CALCDIR, "*_nbo6.out")):
        logger.info(f"Processing {nbo6out}")
        nboobj = nbo.NboCalculation(nbo6out)
        homo_idx, lumo_idx = nbo.get_boundary_indices(nboobj.dmmo.arr)
        mo_energies = np.diag(nboobj.fmo.arr)
        orb_item = {
            'filename': nbo6out,
            # 'HOMO_e': mo_energies[homo_idx],
            'HOMO_idx': int(homo_idx),
            # 'LUMO_e': mo_energies[lumo_idx],
            'LUMO_idx': int(lumo_idx),
        }
        # You can also get MO energies if needed
        for i in range(N_HOMO):
            # orb_item['HOMO-%d_e' % (i + 1)] = mo_energies[homo_idx - i - 1]
            orb_item['HOMO-%d_idx' % (i + 1)] = int(homo_idx - i - 1) # INDEXING STARTS FROM 0
        for i in range(N_LUMO):
            # orb_item['LUMO+%d_e' % (i + 1)] = mo_energies[lumo_idx + i + 1]
            orb_item['LUMO+%d_idx' % (i + 1)] = int(lumo_idx + i + 1)
        orb_data.append(orb_item)
    logger.info("Step 5 is finished")

    logger.info("Step 6. Generate cube-files for the chosen MOs")
    orbital_idxs = []
    mo_assignment = {}
    for orb_item in orb_data:
        molname = ntpath.basename(orb_item['filename']).replace("_nbo6.out", "")
        if molname not in mo_assignment:
            mo_assignment[molname] = {}
        
        cur_mos = []
        for mo_name in MO_IDXS:
            cur_mos.append(orb_item[mo_name + '_idx'])
            mo_assignment[molname][mo_name] = orb_item[mo_name + '_idx']

        mologname = os.path.join(CALCDIR, ntpath.basename(orb_item['filename']).replace("_nbo6.out", "_scf.log"))
        orbital_idxs.append({
            'Logname': mologname,
            'MOs': cur_mos, # Indexing starts with 0
        })
    logger.info(f"Saving MO indices to file {ASSIGNMENT_FNAME}")
    with open(ASSIGNMENT_FNAME, 'w') as f:
        json.dump(mo_assignment, f)
    
    wait_cubes = []
    cube_files = []
    for item in orbital_idxs:
        logger.info("Running cubegen for " + item['Logname'])
        newtasks, newcubefiles = nbo.generate_nbo_cube(item['Logname'], item['Logname'].replace('.log', '.fchk'),
                                                       item['Logname'].replace(".log", "_{nbo}.cube"),
                                                       item['MOs'], gdriver, logger, nbo_reorder=False)
        wait_cubes += newtasks
        cube_files += newcubefiles
    utils.wait_for_termination(wait_cubes, gdriver)
    logger.info("Step 6 is finished")

    logger.info("Step 7. Generate isosurface data")
    for cubefile in glob.glob(os.path.join(CALCDIR, "*.cube")):
        logger.info("Running isosurface calculation for " + cubefile)
        generate_isosurface(cubefile, cubefile.replace('.cube', '_{sign}.{type}iso'), ISOVALUE)
    logger.info("Step 7 is finished")

    logger.info("Step 8. Generate SDFs with topologies from NBO")
    for nbo3name in glob.glob(os.path.join(CALCDIR, "*_nbo3.log")):
        logger.info(f"Processing file = {nbo3name}")
        curmol = geom.Molecule(nbolog=nbo3name)
        curmol.from_xyz(*utils.parse_gaussian_log(nbo3name, input_orientation=True))

        sdfname = nbo3name.replace("_nbo3.log", ".sdf")
        curmol.save_sdf(sdfname)
    logger.info("Step 8 is finished")

    logger.info("My work is done. Terminating.")
    with gdriver['todo_lock']:
        gdriver['todo_files'].append("Done")
