# ./orbitals/nbo.py

import os, sys, copy, glob, ntpath
import shutil, json
from chemscripts.calc import queue
from chemscripts.mylogging import createLogger
from chemscripts.nbo import NBO3LogParser, highest_p_character, generate_nbo_cube, generate_isosurface
from chemscripts import utils, geom

STARTFILE = "./meoome.log" # The name of optimization log-file
CALCDIR = "./calc_files" # Dir for all new files
NBO_MASKS = {
               'donor': r"LP\(.*\)O5$", # Regex for the NBO of interest
               'acceptor': r"BD\*\(1\)C7-H8$", # Another one
            }
ISOVALUE = 0.15 # Isovalue for all NBOs
ASSIGNMENT_FNAME = "./nbo_assignment.json"

if __name__ == "__main__":
    for exe in ("rung", "formchk", "cubegen"):
        assert utils.check_availability(exe), "%s is not included in PATH" % exe

    if len(sys.argv) > 1:
        maxproc = int(sys.argv[1])
    else:
        maxproc = None
    gdriver = queue.init_thread(maxproc)
    logger = createLogger("Main")
    utils.checkout_directory(CALCDIR)
    start_fname = STARTFILE

    logger.info("Step 1. Run NBO3 analysis for initial geometry")
    logger.info("Processing " + start_fname)
    nbo3gjfs = []
    xyzs, syms = utils.parse_gaussian_log(start_fname)
    nbo3_gjfname = os.path.join(CALCDIR, ntpath.basename(start_fname).replace('.log', '_nbo3.gjf'))
    utils.write_gjf(xyzs, syms, 'nbo', nbo3_gjfname,
                    subs={
                        'nproc': 1,
                        'chkname': ntpath.basename(nbo3_gjfname).replace('.gjf', '.chk'),
                        'name': ntpath.basename(nbo3_gjfname).replace('.gjf', ''),
                    })
    nbo3gjfs.append(nbo3_gjfname)
    utils.start_calcs(nbo3gjfs, gdriver)

    for file in glob.glob(os.path.join(CALCDIR, "*.69")):
        shutil.move(file, file.lower().replace(".69", ".fnbo"))
    for file in glob.glob(os.path.join(CALCDIR, "*.420")):
        shutil.move(file, file.lower().replace(".420", ".dmnbo"))
    logger.info("Step 1 is finished")

    logger.info("Step 2. Obtain formatted Gaussian checkpoints")
    fchkcommands = []
    for chkfile in glob.glob(os.path.join(CALCDIR, "*_nbo3.chk")):
        fchkcommands.append(utils.get_formchk_call(chkfile))
    utils.start_calcs(fchkcommands, gdriver)
    logger.info("Step 2 is finished")

    logger.info("Step 3. Obtain indices of NBOs of interest")
    orbitaldata = []
    nbo_assignment = {}
    for logname in glob.glob(os.path.join(CALCDIR, "*_nbo3.log")):
        nboparser = NBO3LogParser(logname)
        nbo_indices = []
        molname = ntpath.basename(logname).replace("_nbo3.log", '')
        if molname not in nbo_assignment:
            nbo_assignment[molname] = {}

        ###### You should customize this part
        for cur_name, cur_regex in NBO_MASKS.items():
            matches = nboparser.find_by_regex(cur_regex)
            assert len(matches) > 0, f"No NBOs found for {cur_regex}"
            if "LP" in cur_regex:
                cur_nbo = highest_p_character(matches)
                logger.warning(f"Using NBO with the highest p-character for {cur_regex}: {cur_nbo['name']}")
            else:
                cur_nbo = matches[0]
                if len(matches) > 1:
                    logger.warning(f"[!!!] Several matches found for {cur_regex}. Using {cur_nbo['name']}")
            nbo_assignment[molname][cur_name] = cur_nbo['index']
            nbo_indices.append(cur_nbo['index'])
        
        orbitaldata.append({
            'Logname': logname,
            'Nbos': nbo_indices, # All NBOs we want to render for 'logname'
        })
        ########
        logger.info(f"[{ntpath.basename(logname)}] Identified these NBO indices: {repr(nbo_indices)}")

    logger.info(f"Saving NBO indices to file {ASSIGNMENT_FNAME}")
    with open(ASSIGNMENT_FNAME, 'w') as f:
        json.dump(nbo_assignment, f)
    logger.info("Step 3 is finished")

    logger.info("Step 4. Generate cube-files")
    wait_tasks = []
    for item in orbitaldata:
        logger.info(f"Generating cube for {item['Logname']}")
        newtask, _ = generate_nbo_cube(item['Logname'], item['Logname'].replace('.log', '.fchk'),
                                        os.path.join(".", item['Logname'].replace(".log", "_{nbo}.cube")),
                                        item['Nbos'], gdriver, logger)
        wait_tasks += newtask
    utils.wait_for_termination(wait_tasks, gdriver)
    logger.info("Step 4 is finished")

    logger.info("Step 5. Generate isosurface data")
    for cubefile in glob.glob(os.path.join(CALCDIR, "*.cube")):
        logger.info("Running isosurface calculation for " + cubefile)
        generate_isosurface(cubefile, cubefile.replace('.cube', '_{sign}.{type}iso'), ISOVALUE)
    logger.info("Step 5 is finished")

    logger.info("Step 6. Generate SDFs with topologies from NBO")
    for nbo3name in glob.glob(os.path.join(CALCDIR, "*_nbo3.log")):
        logger.info(f"Processing file = {nbo3name}")
        curmol = geom.Molecule(nbolog=nbo3name)
        curmol.from_xyz(*utils.parse_gaussian_log(nbo3name, input_orientation=True))

        sdfname = nbo3name.replace("_nbo3.log", ".sdf")
        curmol.save_sdf(sdfname)
    logger.info("Step 6 is finished")

    logger.info("My work is done. Terminating.")
    with gdriver['todo_lock']:
        gdriver['todo_files'].append("Done")
